/**************************************************************************/
/* Title:       vramc                                                     */
/* Date:        Tuesday 8th September 1992                                */
/*------------------------------------------------------------------------*/
/* Version: Amended:  By:            Changes:                             */
/*------------------------------------------------------------------------*/
/* 0.1       8/9/92   Bill Oldfield   started                             */
/* 0.2      10/9/92   Bill Oldfield   basic model working                 */
/* 0.3      10/9/92   Bill Oldfield   added dsf and qsf pins              */
/* 0.4      11/9/92   Bill Oldfield   split transfers working             */
/* 0.5      11/9/92   Bill Oldfield   added rows and cols as sidle input  */
/* 0.6      23/9/92   Bill            correected rising CAS in RASBEFORERA*/
/* 0.7     26/11/92   Bill            made sio goto XXX on Nsoe and sc    */
/* 0.8     18/02/93   Mark Taunton    fixed pre-initialisation from file  */
/* 0.9     04/03/93   Mark Taunton    ensure wrap-around of SAM counter   */
/**************************************************************************/
/* supports Ncas before Nras refresh, early/late write, read, read -      */
/* transfer and split-read transfer cycles. Will produce a warning if it  */
/* doesn't understand a cycle. At present it does no timing checking on   */
/* inputs. The maximum width of the vram is 8-bit, since an array of chars*/
/* is used as the storage. The array could be implemented using a hash    */
/* table if memory is a problem and the vram is not usually filled. Needs */
/* the following parameters:                                              */
/*           tCAC  falling Ncas  to data out                              */
/*           tOFF1  rising Ncas  to data tri-state                        */
/*           tOFF2  rising Ndtoe to data tri-state                        */
/*           tSEZ   rising Nsoe  to sio  tri-state                        */
/*           tSEA  falling Nsoe  to sio  out                              */
/*           tSCA   rising sc    to sio  out                              */
/*           tSOH   rising sc    to sio  tri-state                        */
/*           tPD    rising sc    to qsf  out                              */
/*           tDCD   rising Nras (transfer cycle) to qsf out               */
/*           rows  number of rows in vram memory array                    */
/*           cols  number of columns in vram memory array                 */
/**************************************************************************/

#include <Handy.h>

#define        UNKNOWN 0
#define CAS_BEFORE_RAS 1
#define       TRANSFER 2
#define SPLIT_TRANSFER 3
#define            RAM 4

#define address              (state->coladd + (state->rowadd<<state->cols))
#define colmsb(counter,cols) (1&(counter>>(cols-1)))
#define colbits(cols)        (1<<(cols))       /* eg colbits=256 for cols=8 */
#define colmask(cols)        ((1<<(cols))-1)   /* eg colmask=255 for cols=8 */
#define colhmask(cols)       ((1<<((cols)-1))-1) /* mask for half a row (split SAM) */

/*-----------------------VRAM Model Definition----------------------------*/

#define VRAMC_def  "VRAMC(Nras,Ncas,Nwe,Ndtoe,ra[],Nsoe,dsf,sc,data[])=>(sio[],qsf)"

typedef struct {
  ModelCore model;
  Signal *Nras, *Ncas, *Nwe, *Ndtoe;
  Bundle *ra;
  Signal *Nsoe, *dsf, *sc;
  Bundle *data, *sio;
  Signal *qsf;
} ModelOfVRAMC;

typedef struct {
  unsigned char *vram_array, *sam_array;
  unsigned sam_counter;
  unsigned rowadd, coladd;
  unsigned tCAC,tOFF1,tOFF2,tSEZ,tSEA,tSCA,tSOH,tPD,tDCD;
  unsigned access_type,rows,cols;
} VRAMCState;

/*------------- Execute transfer of ram data to sam data --------------*/

void do_transfer(ram_array,sam_array,rows,cols,type,current,rowadd,coladd)
unsigned char *ram_array,*sam_array;
unsigned rows,cols,type,current,rowadd,coladd;
{
   int i;

   if (type==TRANSFER) {                      /* whole transfer */
      for (i=0; i<colbits(cols); i++) {
         sam_array[i] = ram_array[ i + (rowadd<<rows) ];
      }
   }
   else if (type==SPLIT_TRANSFER) {
      if (colmsb(coladd,cols)) {              /*  top half transfer  */
         for (i=colbits(cols)/2; i<colbits(cols); i++) {
            sam_array[i] = ram_array[ i + (rowadd<<rows) ];
         }
      }
      else {
         for (i=0; i<colbits(cols)/2; i++) {  /*  bottom half transfer  */
            sam_array[i] = ram_array[ i + (rowadd<<rows) ];
         }
      }
   }
}

/*--------------------VRAM Model Code-------------------------------------*/

static void VRAMC_proc(ModelInst)
VoidStar ModelInst;
{
  ModelOfVRAMC *me = (ModelOfVRAMC *) ModelInst;
  VRAMCState *state = me->model.state;
  int i;

/*--- The state following bit ------*/

  if (FALLING(Nras)) {
     state->rowadd = BINVAL(ra);    /* Nras clocks in row address */
     if (state->access_type==UNKNOWN) {
        if      (LO(Ndtoe) && LO(dsf)) state->access_type = TRANSFER;
        else if (LO(Ndtoe) && HI(dsf)) state->access_type = SPLIT_TRANSFER;
        else if (HI(Ndtoe))            state->access_type = RAM;
     }
  }

/*--- The parallel host interface ----*/

  if (FALLING(Ncas)) {
     state->coladd = BINVAL(ra);   /* Ncas clocks in the column address  */
     if (HI(Nras)) state->access_type = CAS_BEFORE_RAS;
     else if (state->access_type==RAM) {  /* read cycle */
        if (HI(Nwe)&&LO(Ndtoe)) BundleEvt(data, state->vram_array[address],state->tCAC);
        else if (LO(Nwe)){            /* early write */
           state->vram_array[address] = BINVAL(data);
        }
     }
     else if (state->access_type==TRANSFER);
     else if (state->access_type==SPLIT_TRANSFER);
     else errhandler_f2(SS_MODELS+WARNING,"%s: Unknown VRAM access type (%d)\n",(int)INSTNAME,
           state->access_type);
  }

  if (RISING(Ncas)) {
     BundleZZZ(data, state->tOFF1);
     if (state->access_type==CAS_BEFORE_RAS) state->access_type=UNKNOWN;
  }

  if (FALLING(Nwe)) {
     if (LO(Ncas)) {      /* delayed write */
        state->vram_array[address] = BINVAL(data);
     }
  }

  if (RISING(Ndtoe)) {  /* Nras OR Ndtoe trigger transfer whichever is first */
     BundleZZZ(data, state->tOFF2);
     if ((state->access_type==TRANSFER) || (state->access_type==SPLIT_TRANSFER)) {
        do_transfer(state->vram_array, state->sam_array, state->rows,state->cols,
                    state->access_type, state->sam_counter, state->rowadd, state->coladd);
        if (state->access_type==TRANSFER) {
           state->sam_counter = state->coladd;
           if (colmsb(state->sam_counter,state->cols)) SignalEvt(qsf, ONE,state->tDCD);
           else                                        SignalEvt(qsf,ZERO,state->tDCD);
        }
     }
     if (HI(Nras)) state->access_type = UNKNOWN;
  }

  if (RISING(Nras)) {
     if ((state->access_type==TRANSFER) || (state->access_type==SPLIT_TRANSFER)) {
        do_transfer(state->vram_array, state->sam_array, state->rows,state->cols,
                    state->access_type, state->sam_counter, state->rowadd, state->coladd);
        if (state->access_type==TRANSFER) {
           state->sam_counter = state->coladd;
           if (colmsb(state->sam_counter,state->cols)) SignalEvt(qsf, ONE,state->tDCD);
           else                                        SignalEvt(qsf,ZERO,state->tDCD);
        }
     }
     state->access_type = UNKNOWN;
  }

/*--- the sam port model -----*/

  if (RISING(Nsoe))  BundleZZZ(sio, state->tSEZ);
  if (FALLING(Nsoe)) {
     BundleXXX(sio, state->tSOH);
     BundleEvt(sio, state->sam_array[state->sam_counter], state->tSEA);
  }

  if (RISING(sc)) {
     BundleXXX(sio, state->tSOH);   /* changed from ZZZ 26/11/92 by request of Ashley */
     if (LO(Nsoe)) BundleEvt(sio, state->sam_array[state->sam_counter], state->tSCA);
     state->sam_counter = (state->sam_counter+1) & colmask(state->cols);
     if ((state->sam_counter & colhmask(state->cols)) == 0) {
	/* msb has just toggled: toggle qsf to reflect this */
	if (colmsb(state->sam_counter,state->cols)) SignalEvt(qsf, ONE, state->tPD);
	else                                        SignalEvt(qsf,ZERO, state->tPD);
     }
  }
}
  
/*-----------------VRAM Initialisation-----------------------------*/

static void VRAMC_init(ModelInst)
VoidStar ModelInst;
{
  VRAMCState *s;
  ModelOfVRAMC *me = (ModelOfVRAMC* ) ModelInst;
  Parameter param;
  unsigned size;
  char filename[255];
  FILE *file_ptr;
  NEW(s, VRAMCState);

  me->model.proc = VRAMC_proc;             

  param = simmodel_param(me->model.state, "rows"); /* number of rows in the matrix */
  if (simmodel_type(param) != PT_INTEGER)
      errhandler_f1(SS_MODELS+DISASTER, "%s: VRAM 'rows' is NOT properly specified\n", (int)INSTNAME);
  s->rows = simmodel_int(param);

  param = simmodel_param(me->model.state, "cols"); /* number of columns in the matrix */
  if (simmodel_type(param) != PT_INTEGER) 
      errhandler_f1(SS_MODELS+DISASTER, "%s: VRAM 'cols' is NOT properly specified\n", (int)INSTNAME);
  s->cols = simmodel_int(param);

  if ((s->rows+s->cols)>26)
      errhandler_f1(SS_MODELS+DISASTER, "%s: VRAM size too large\n",(int)INSTNAME);
  size = (1<<s->rows) * (colbits(s->cols));
  s->vram_array = (unsigned char *)malloc (size);
/*  s->vram_array = (unsigned char *)malloc ((1<<s->rows) * (colbits(s->cols))); */
  if (s->vram_array==0)
      errhandler_f1(SS_MODELS+DISASTER, "%s: Not enough memory for storage\n", (int)INSTNAME);

  s->sam_array = (unsigned char *)malloc(colbits(s->cols));
  if (s->sam_array==0)
      errhandler_f1(SS_MODELS+DISASTER,"%s: Not enough memory for sam buffer\n",(int)INSTNAME);

  s->tCAC  = DELAY("tCAC");   /* falling Ncas  to data out                */
  s->tOFF1 = DELAY("tOFF1");  /*  rising Ncas  to data tri-state          */
  s->tOFF2 = DELAY("tOFF2");  /*  rising Ndtoe to data tri-state          */
  s->tSEZ  = DELAY("tSEZ");   /*  rising Nsoe  to sio  tri-state          */
  s->tSEA  = DELAY("tSEA");   /* falling Nsoe  to sio  out                */
  s->tSCA  = DELAY("tSCA");   /*  rising sc    to sio  out                */
  s->tSOH  = DELAY("tSOH");   /*  rising sc    to sio  tri-state          */
  s->tPD   = DELAY("tPD");    /*  rising sc    to qsf  out                */
  s->tDCD  = DELAY("tDCD");   /*  rising Nras (transfer cycle) to qsf out */

  s->access_type = UNKNOWN;

  s->sam_counter=0;

  /* Check for optional pre-initialisation.  Reworked by Mark 17-Feb-93. */
  param = simmodel_param(me->model.state, "filename");
  if (simmodel_type(param)!=PT_UNDEFINED) {
    int groupsize, byte;
    int bno, vno, boff;
    unsigned char buff[BUFSIZ];

    strcpy(filename, simmodel_str(param));
    /*
     * Determine number of byte-wide VRAM chips in use (1..16): if
     * more than 16, the loading algorithm here becomes stupidly
     * inefficient and it would be better to have separate files, one
     * for each VRAM.
     */
    param = simmodel_param(me->model.state, "groupsize");
    if (simmodel_type(param) != PT_INTEGER || 
	(groupsize = simmodel_int(param)) < 1 || groupsize > 16)
	errhandler_f1(SS_MODELS+DISASTER,
		      "%s: VRAM 'groupsize' is not properly specified\n", (int)INSTNAME);

    /* Determine which byte of the group this VRAM instance represents. */
    param = simmodel_param(me->model.state, "byte");
    if (simmodel_type(param) != PT_INTEGER || 
	(byte = simmodel_int(param)) < 0 || byte >= groupsize)
	errhandler_f1(SS_MODELS+DISASTER,
		      "%s: VRAM 'byte' is not properly specified\n", (int)INSTNAME);

    /* Go open the file */
    file_ptr = simfileio_find_input(NULL,filename);       
    if (file_ptr==NULL)
	errhandler_f (SS_MODELS+DISASTER,
		      "%s: vram 'filename'=%s, not found\n", (int)INSTNAME, (int)filename);

    /*
     * Load up data from the file, using only the 'byte'th byte of
     * each 'groupsize'-sized group from the file, but initialising
     * each consecutive byte of the VRAM array.  Note that we cope
     * with arbitary group sizes from 1..16, and are not limited to
     * powers of 2.
     */
    vno = 0;  boff = byte;
    for (;;)
    {
	int bc, bn;
	bc = fread (buff, 1, sizeof(buff), file_ptr);
	if (bc == 0 || vno == size)
	    break;
	for (bn = boff; bn < bc && vno < size; bn += groupsize)
	    s->vram_array[vno++] = buff[bn];

	boff = bn % bc;			/* offset into next buffer for next time */
    }
    errhandler_f (SS_MODELS+OK,
		  "%s: vram byte %d/%d pre-initialised from 'filename'=%s\n",
		  (int)INSTNAME, byte, groupsize, (int)filename);
    fclose(file_ptr);
  }

  SignalEvt(qsf, ZERO, 1);
  BundleZZZ(data, 1);
  BundleZZZ( sio, 1);

  me->model.state = (VoidStar ) s;
}

/*****************************Model Declaration****************************/

void VRAMC_declare()
{
  simmodel_begin("iomd.vramc");
  simmodel_declare_s(VRAMC_def, VRAMC_init);
  simmodel_end();
}
