/* h.handy

  Model-writers' interface to the exported functions of the Asim kernel

  Includes useful #defined macros which makes the assumption that:
    variable me is a pointer to a struct containing the model instance information
    including its signals and bundles (type ModelInstance, defined in simsimtyps.h)

Version
  0.1   14-06-89  C Garlick
  0.2   11-09-89  Jim Sutton
  0.3   10/10/89  J Sutton changed macros:
                    SignalEvt(sig, val, del)  does NOT bitwise mask the value
                    BundleXXX(bund, del)      puts 0 into value, NOT the current value
                                              The current value is irrelevant for future
                                              scheduling
                    BundleZZZ(bund, del)      ditto, and the value is not used when
                                              it is descheduled anyway.
                  removed macros SignalXXX and SignalZZZ, as the functions were
                  misleading and not useful.
  0.4   04/12/90  A Merritt added BXX and BZZ macros for Bundle use
                  Also added comments to some of the more misusable Macros
                  Added a cast to unsigned in macro BM to stop warnings if n=31
  1.0   20/09/91  A Merritt changed many macros to upper case to allow the Sun
                  compiler to function - all models will have to be updated before
                  using this new Handy.h  See me for details.        
  1.1   07/01/92  A Thomas added the SignalZZZ macro to allow a signal's drive to be
                  removed *without* affecting its current level (this is *NOT* possible
                  with the SignalEvt macro!)
*/

#ifndef _def__Handy
#define _def__Handy

#include "simmodel.h"
#include "simsimtyps.h"
#include "simbastyps.h"
#include "simkernel0.h"
#include "simkernel1.h"
#include "simfileio.h"
#include "simq.h"
#include "stdtypes.h"
#include "stddefs.h"
#include "streams.h"
#include "errhandler.h"
#include "errorcodes.h"
#include "morestring.h"
#include "bigt.h"
#include "gettable.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

/* The state bits are:

    ZERO        = 0
    ONE         = 1
    UNDEFINED   = 2
    X           = UNDEFINED
    HIZ         = 4
    SET_BY_CLI  = 16

defined in simsimtyps.h
*/

#ifndef TRUE
#define TRUE   1 
#define FALSE  0
#endif

#define ALLONES       0xffffffff
#define DEF           0 /* Defined */
#define BM(n)         ((((unsigned) 1)<<(n))-1) /* Bit Mask */

/**********************************************************************************************************/

/* The following Macros should be used for Bundles only */
/* Note that bundles must be 'gathered' before their first use, so only use bval if you are sure of yourself */
#define GATHER(b)     simkernel1_gather(me->b)
#define BINVAL(b)     (((FB_OUTOFDATE&me->b->flags) ? simkernel1_gather(me->b),0:0), me->b->val.binValue)
#define BVAL(b)       (me->b->val.binValue)
#define BZZ(b)        (((FB_OUTOFDATE&me->b->flags) ? simkernel1_gather(me->b),0:0),((HIZ&me->b->value) ? TRUE : FALSE))
#define BXX(b)        (((FB_OUTOFDATE&me->b->flags) ? simkernel1_gather(me->b),0:0),((X&me->b->value) ? TRUE : FALSE))

/* The following Macros should be used for Signals only */
#define LVL(s)        (ONE&me->s->value) /* Logic level for equating signals */
#define HI(s)         ((ONE&me->s->value) ? TRUE : FALSE) /* ANY High : undefined or tristate */ 
#define LO(s)         ((ONE&~me->s->value) ? TRUE : FALSE) /* ANY Low : undefined or tristate */
#define ZZZ(s)        ((HIZ&me->s->value) ? TRUE : FALSE) 
#define XXX(s)        ((X&me->s->value) ? TRUE : FALSE)
#define RISING(s)     (((FB_CHANGED&me->s->flags)&&(ONE&me->s->value)) ? TRUE : FALSE)
#define FALLING(s)    (((FB_CHANGED&me->s->flags)&&(ONE&~me->s->value)) ? TRUE : FALSE)
#define DHI(s)        (((X|ONE)&me->s->value)==ONE) /* Defined High : Tristate or not */
#define DLO(s)        (((X|ONE)&me->s->value)==ZERO) /* Defined Low : Tristate or not */ 

/* The following Macros may be used with either Signals or Bundles */
/* Note that val gives the combined status and value of a Signal - this use is fine */
/* But that it returns the state only for a Bundle, and that it does not gather the Bundle - use with care */
#define VAL(s)        (me->s->value)
#define IN(numb,bit)  ((1&((numb)>>(bit))) ? TRUE : FALSE) /* logical evaluation of a single-bit */
#define CHANGING(s)   ((FB_CHANGED&me->s->flags) ? TRUE : FALSE)

/**********************************************************************************************************/

#define SignalEvt(sig, val, del)  simq_signal_evt(me->sig, val, me, del)
/* SignalEvt sets a value and state on a signal after a specified time delay
 *
 * NOTE: If this macro is used to send a signal HIZ, the signal level will
 * ####  also be set to a new value as the drive is removed.
 *       The SignalZZZ macro should be used to remove a signal's drive
 *       without affecting its level */

#define SignalZZZ(sig, del)       simq_signal_zzz(me->sig, me, del)
/* SignalZZZ will mark the signal as undriven without affecting its current level */

#define BundleEvt(bund, val, del) simq_bundle_evt(me->bund, DEF, val, me, del)
/* BundleEvt sets a value (only) on a bundle after a specified time delay */

#define BundleXXX(bund, del)      simq_bundle_evt(me->bund, X, 0, me, del)
/* BundleXXX will set the bundle to Undefined Zero */

#define BundleZZZ(bund, del)      simq_bundle_evt(me->bund, HIZ, 0, me, del)
/* BundleZZZ will mark the bundle as undriven, but does not affect its binary value when
 * it is descheduled */
/* N.B. This trick happens in module simkernel1, and only affects bundles, not signals */

#define RemoveEvt(sb, bigt)       ((bigt_comp(bigt, &simq_now)==COMPARISON_GT) ? simq_bigt_remove_evt(me->sb, bigt) : TRUE)
/* remove a previously-scheduled event from the queue.
 * Returns FALSE if it failed to find the event, else TRUE if the time is past, OR it removed it */

#define INSTNAME                  (simkernel0_instname(&me->model))
#define DELAY(name)               (simmodel_delay(name, &me->model))

/* set param to the parameter in the SIDLE tagged with str. Return the type of the param */
#define get_param(param, str) (simmodel_type((param)=simmodel_param(me->model.state, str)))

/* case-independent string comparison */
#define cistrcmp(param, str) (morestring_ci_comp(simmodel_str(param), str) == COMPARISON_EQ)

/* compare times where bigt is type bigt_Time, and t is int */
#define time_comp(bigt, t) \
   ( ((bigt)->tHi)       ? COMPARISON_GT : ((bigt)->tLo > (t)) ? COMPARISON_GT : \
     ((bigt)->tLo < (t)) ? COMPARISON_LT : \
                           COMPARISON_EQ )

#endif
